﻿using System;
using System.IO;
using System.Drawing;

namespace Sat2Ip.Library.Rtp
{
    public class RtpPacket
    {
        private const bool DebugHeader = false;
        private const int HeaderSize = 12;
        public int Version { get; set; }
        public int Padding { get; set; }
        public int Extension { get; set; }
        public int Cc { get; set; }
        public int Marker { get; set; }
        public int PayloadType { get; set; }
        public int SequenceNumber { get; set; }
        public int TimeStamp { get; set; }
        public int Ssrc { get; set; }
        public byte[] Header { get; set; }
        public int PayloadSize { get; set; }
        public byte[] Payload { get; set; }

        //--------------------------
        //Constructor of an RTPpacket object from header fields and payload bitstream
        //--------------------------
        public RtpPacket(int pType, int framenb, int time, byte[] data, int dataLength)
        {
            //fill by default header fields:
            Version = 2;
            Padding = 0;
            Extension = 0;
            Cc = 0;
            Marker = 0;
            Ssrc = 0;

            //fill changing header fields:
            SequenceNumber = framenb;
            TimeStamp = time;
            PayloadType = pType;

            //build the header bistream:
            //--------------------------
            Header = new byte[HeaderSize];

            //fill the header array of byte with RTP header fields
            Header[0] = (byte)((Version << 6) | (Padding << 5) | (Extension << 4) | Cc);
            Header[1] = (byte)((Marker << 7) | PayloadType);

            Header[2] = (byte)(SequenceNumber >> 8);
            Header[3] = (byte)(SequenceNumber);
            for (int i = 0; i < 4; i++)
                Header[7 - i] = (byte)(TimeStamp >> (8 * i));

            for (int i = 0; i < 4; i++)
                Header[11 - i] = (byte)(Ssrc >> (8 * i));

            //fill the payload bitstream:
            //--------------------------
            PayloadSize = dataLength;
            Payload = new byte[dataLength];

            //fill payload array of byte from data (given in parameter of the constructor)
            for (int i = 0; i < dataLength; i++)
                Payload[i] = data[i];
        }

        //--------------------------
        //Constructor of an RTPpacket object from the packet bistream 
        //--------------------------
        public RtpPacket(byte[] packet, int packetSize)
        {
            //fill default fields:
            Version = GetRtpHeaderValue(packet, 0, 1);
            Padding = GetRtpHeaderValue(packet, 2, 2); ;
            Extension = GetRtpHeaderValue(packet, 3, 3); ;
            Cc = GetRtpHeaderValue(packet, 4, 7);
            Marker = GetRtpHeaderValue(packet, 8, 8); ;
            Ssrc = GetRtpHeaderValue(packet, 64, 95); ;

            //check if total packet size is lower than the header size
            if (packetSize >= HeaderSize)
            {
                //get the header bitsream:
                Header = new byte[HeaderSize];
                for (int i = 0; i < HeaderSize; i++)
                    Header[i] = packet[i];

                //get the payload bitstream:
                PayloadSize = packetSize - HeaderSize;
                Payload = new byte[PayloadSize];
                for (int i = HeaderSize; i < packetSize; i++)
                    Payload[i - HeaderSize] = packet[i];

                //interpret the changing fields of the header :
                PayloadType = Header[1] & 127;
                SequenceNumber = unsigned_int(Header[3]) + 256 * unsigned_int(Header[2]);
                TimeStamp = unsigned_int(Header[7]) + 256 * unsigned_int(Header[6]) + 65536 * unsigned_int(Header[5]) + 16777216 * unsigned_int(Header[4]);
            }
        }

        //--------------------------
        //getpayload : return the payload bistream of the RTPpacket and its size
        //--------------------------
        public int GetPayload(byte[] data)
        {

            for (int i = 0; i < PayloadSize; i++)
                data[i] = Payload[i];

            return (PayloadSize);
        }

        //--------------------------
        //getpayload_length : return the length of the payload
        //--------------------------
        public int GetPayloadLength()
        {
            return (PayloadSize);
        }

        //--------------------------
        //getlength : return the total length of the RTP packet
        //--------------------------
        public int GetLength()
        {
            return (PayloadSize + HeaderSize);
        }

        //--------------------------
        //getpacket : returns the packet bitstream and its length
        //--------------------------
        public int GetPacket(byte[] packet)
        {
            for (var i = 0; i < HeaderSize; i++)
                packet[i] = Header[i];
            for (var i = 0; i < PayloadSize; i++)
                packet[i + HeaderSize] = Payload[i];
            return (PayloadSize + HeaderSize);
        }

        //--------------------------
        //gettimestamp
        //--------------------------

        public int GetTimestamp()
        {
            return (TimeStamp);
        }

        //--------------------------
        //getsequencenumber
        //--------------------------
        public int GetSequencenumber()
        {
            return (SequenceNumber);
        }

        //--------------------------
        //getpayloadtype
        //--------------------------
        public int GetPayloadType()
        {
            return (PayloadType);
        }


        //--------------------------
        //print headers without the SSRC
        //--------------------------
        public void PrintHeader()
        {
            if (DebugHeader)
            {
                for (int i = 0; i < (HeaderSize - 4); i++)
                {
                    for (int j = 7; j >= 0; j--)
                        Console.Out.Write(((1 << j) & Header[i]) != 0 ? "1" : "0");
                    Console.Out.Write(" ");
                }
                Console.Out.WriteLine();
            }
        }

        //return the unsigned value of 8-bit integer nb
        static int unsigned_int(int nb)
        {
            if (nb >= 0)
                return (nb);
            return (256 + nb);
        }

        public Image GetPayloadAsImage()
        {
            var payloadLength = GetPayloadLength();
            var payload = new byte[payloadLength];
            GetPayload(payload);
            Stream imageInMemory = new MemoryStream(payload);
            return Image.FromStream(imageInMemory);
        }

        public RtpStream GetPayloadAsTransportStream()
        {
            var payloadLength = GetPayloadLength();
            var payload = new byte[payloadLength];
            GetPayload(payload);
            RtpStream stream = new RtpStream(payload.Length);
            return stream;
        }
        private int GetRtpHeaderValue(byte[] packet, int startBit, int endBit)
        {
            int result = 0;

            // Number of bits in value
            int length = endBit - startBit + 1;

            // Values in RTP header are big endian, so need to do these conversions
            for (int i = startBit; i <= endBit; i++)
            {
                int byteIndex = i / 8;
                int bitShift = 7 - (i % 8);
                result += ((packet[byteIndex] >> bitShift) & 1) * (int)Math.Pow(2, length - i + startBit - 1);
            }
            return result;
        }
    }
}


