﻿using System;
using System.ComponentModel;
using System.Globalization;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Text.RegularExpressions;


namespace Sat2Ip.Library.Rtsp
{
    public class RtspSession : INotifyPropertyChanged ,IDisposable
    {
        #region Private Fields

        private RtspDevice _rtspDevice;
        private Socket _rtspSocket;
        private int _rtspSequenceNum = 0;
        private string _rtspResponse;
        private string _rtspSessionId;
        private int _rtspSessionTimeToLive;
        private string _rtspStreamId;
        private string _serverRtpPort;
        private string _serverRtcpPort;
        private string _rtspStreamUrl;
        private string _destination;
        private string _source;
        private string _transport;

        #endregion

        #region Constructor

        public RtspSession(RtspDevice rtspDevice)
        {
            _rtspDevice = rtspDevice;
            
        }

        #endregion

        #region Properties
        public string Transport
        {
            get
            {
                if (string.IsNullOrEmpty(_transport))
                {
                    _transport = "unicast";
                }
                return _transport;
            }
            set 
            {
                if (_transport != value)
                {
                    _transport = value;
                    OnPropertyChanged("Transport");
                }
            }
        }
        public string Destination
        {
            get
            {
                if (string.IsNullOrEmpty(_destination))
                {
                    var host = Utils.GetLocalHostName();
                    _destination = Utils.GetIpAddress(host);
                }
                return _destination;
            }
            set
            {
                if (_destination != value)
                {
                    _destination = value;
                    OnPropertyChanged("Destination");
                }
            }
        }
        public string Source
        {
            get { return _source; }
            set
            {
                if (_source != value)
                {
                    _source = value;
                    OnPropertyChanged("Source");
                }
            }
        }
        public string ServerRtpPort
        {
            get
            {
                return _serverRtpPort; 
            }
            set { if (_serverRtpPort != value) { _serverRtpPort = value; OnPropertyChanged("ServerRtpPort"); } }
        }
        public string ServerRtcpPort
        {
            get
            {
                return _serverRtcpPort;
            }
            set { if (_serverRtcpPort != value) { _serverRtcpPort = value; OnPropertyChanged("ServerRtcpPort"); } }
        }
        public string RtspStreamId
        {
            get { return _rtspStreamId; }
            set { if (_rtspStreamId != value) { _rtspStreamId = value; OnPropertyChanged("RtspStreamId"); } }
        }
        public string RtspStreamUrl
        {
            get { return _rtspStreamUrl; }
            set { if (_rtspStreamUrl != value) { _rtspStreamUrl = value; OnPropertyChanged("RtspStreamUrl"); } }
        }
        public string RtspResponse
        {
            get { return _rtspResponse; }
            set { if(_rtspResponse!=value){_rtspResponse = value; OnPropertyChanged("");} }
        }
        public RtspDevice RtspDevice
        {
            get { return _rtspDevice; }
            set { if (_rtspDevice != value) { _rtspDevice = value; OnPropertyChanged("RtspDevice"); } }
        }
        public int RtspSessionTimeToLive
        {
            get { return _rtspSessionTimeToLive; }
            set { if (_rtspSessionTimeToLive != value) { _rtspSessionTimeToLive = value; OnPropertyChanged("RtspSessionTimeToLive"); } }
        }

        #endregion

        #region Private Methods

        private void Connect()
        {
            _rtspSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            var ip = IPAddress.Parse(RtspDevice.RtspServerAddress);
            var port = Convert.ToInt16(RtspDevice.RtspServerPort);
            var rtspEndpoint = new IPEndPoint(ip, port);
            _rtspSocket.Connect(rtspEndpoint);
        }
        private void Disconnect()
        {
            if (_rtspSocket != null && _rtspSocket.Connected)
            {
                _rtspSocket.Shutdown(SocketShutdown.Both);
                _rtspSocket.Close();
            }
        }
        private bool SendRequest(string request)
        {
            var result = false;
            if (_rtspSocket == null)
            {

            }
            try
            {
                var sendBuffer = Utils.StringToBytes(request);
                var sendBytesCount = _rtspSocket.Send(sendBuffer, sendBuffer.Length, SocketFlags.None);
                if (sendBytesCount < 1)
                {
                    result = false;
                }

                result = true;
            }
            catch (Exception e)
            {
                result = false;
            }
            return result;

        }
        private bool GetResponse()
        {
            var result = false;
            var revBytesCount = 0;
            var revBuffer = new byte[1024 * 4];
            RtspResponse = string.Empty;
            _rtspSocket.ReceiveTimeout = 5000;
            while (true)
            {
                try
                {
                    revBytesCount = _rtspSocket.Receive(revBuffer, revBuffer.Length, SocketFlags.None);
                    if (revBytesCount >= 1)
                    {
                        break;
                    }
                    if (_rtspSocket.ReceiveTimeout >= 5000)
                    {
                        break;
                    }
                }
                catch (SocketException)
                {

                }
            }
            if (revBytesCount >= 1)
            {
                RtspResponse = Utils.BytesToString(revBuffer, revBytesCount);
                if (RtspResponse.StartsWith("RTSP/1.0"))
                {
                    var dstArray = RtspResponse.Split(' ');
                    if (dstArray.Length > 1)
                    {
                        var code = dstArray[1];
                        if (code.Equals("200"))
                        {
                            result = true;
                        }
                    }
                }
            }
            return result;
        }

        #endregion

        #region Public Methods

        public bool Setup(string query)
        {
            var result = false;
            if ((_rtspSocket==null))
            {
                Connect();
            }
            var sb = new StringBuilder();
            if (string.IsNullOrEmpty(_rtspSessionId))
            {
                sb.AppendFormat("{0} ", "SETUP");
                sb.AppendFormat("rtsp://{0}:{1}/?{2} RTSP/1.0\r\n", RtspDevice.RtspServerAddress,RtspDevice.RtspServerPort, query);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                sb.AppendFormat("Transport: RTP/AVP;unicast;client_port={0}-{1}\r\n\r\n",  RtspDevice.RtpClientPort, RtspDevice.RtcpClientPort);
            }
            else
            {
                sb.AppendFormat("{0} ", "SETUP");
                sb.AppendFormat("rtsp://{0}:{1}/?{2} RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort, query);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                sb.AppendFormat("Session: {0}\r\n", _rtspSessionId);
                sb.AppendFormat("Transport: RTP/AVP;unicast;client_port={0}-{1}\r\n\r\n",  RtspDevice.RtpClientPort, RtspDevice.RtcpClientPort);
            }
            result = SendRequest(sb.ToString());
            result = GetResponse();
            var lineArray = Regex.Split(RtspResponse, @"[\r\n]+");
            foreach (var s in lineArray)
            {
                if(s.Contains("Date:"))
                {}
                if (s.Contains("Session:"))
                {
                    var sessionline = s;
                    if (string.IsNullOrEmpty(sessionline))
                    {

                    }
                    var sessionlinevalues = sessionline.Split(';');
                    var sessionvalues = sessionlinevalues[0].Split(':');
                    _rtspSessionId = sessionvalues[1];
                    if(sessionlinevalues.Length> 1)
                    {
                        var ttlvalues = sessionlinevalues[1].Split('=');
                        _rtspSessionTimeToLive = int.Parse(ttlvalues[1]);
                    }
                    
                }
                if (s.Contains("Transport:"))
                {
                    var transportline = s;
                    if (string.IsNullOrEmpty(transportline))
                    {

                    }
                    var transportvalues = transportline.Split(';');
                    foreach (var transportvalue in transportvalues)
                    {
                        if (transportvalue.Contains("server_port"))
                        {
                            var serverportvalues = transportvalue.Substring(("server_port").Length)
                                                                 .Trim()
                                                                 .Split('=');

                            var ports = serverportvalues[1].Split('-');
                            ServerRtpPort = ports[0];
                            ServerRtcpPort = ports[1];
                        }
                        if (transportvalue.Contains("destination"))
                        {
                            var destinationvalues = transportvalue.Substring(("destination").Length)
                                                                 .Trim()
                                                                 .Split('=');
                            Destination = destinationvalues[1];
                        }
                        if (transportvalue.Contains("source"))
                        {

                        }
                    }
                }
                if (s.Contains("com.ses.streamID:"))
                {
                    RtspStreamId = s.Substring(("com.ses.streamID:").Length).Trim();
                }
            }

            return result;
        }
        public bool Play(string query)
        {
            var result = false;
            if ((_rtspSocket == null) )
            {
                Connect();
            }
            var sb = new StringBuilder();
            if (string.IsNullOrEmpty(query))
            {
                sb.AppendFormat("{0} ", "PLAY");
                sb.AppendFormat("rtsp://{0}:{1}/stream={2} RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort,
                                RtspStreamId);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                
                sb.AppendFormat("Session: {0}\r\n\r\n", _rtspSessionId);

            }
            else
            {
                sb.AppendFormat("{0} ", "PLAY");
                sb.AppendFormat("rtsp://{0}:{1}/stream={2}?{3} RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort,
                                RtspStreamId, query);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                
                sb.AppendFormat("Session: {0}\r\n\r\n", _rtspSessionId);
            }
            result = SendRequest(sb.ToString());
            result = GetResponse();
            var lineArray = Regex.Split(RtspResponse, @"[\r\n]+");
            foreach (var s in lineArray)
            {
                if (s.Contains("Session:"))
                {
                    var sessionline = s;
                    if (string.IsNullOrEmpty(sessionline))
                    {

                    }
                    var sessionvalues = sessionline.Split(':');
                    _rtspSessionId = sessionvalues[1];

                }
                if (s.Contains("RTP-Info:"))
                {
                    var rtpinfoline = s;
                    if (string.IsNullOrEmpty(rtpinfoline))
                    {

                    }

                    var rtpinfovalues = rtpinfoline.Split(' ');
                    RtspStreamUrl = rtpinfovalues[1].Replace("url=", "");
                }
            }
            return result;
        }
        public bool Options()
        {
            if ((_rtspSocket == null) )
            {
                Connect();
            }
            var sb = new StringBuilder();
            if (string.IsNullOrEmpty(_rtspSessionId))
            {
                sb.AppendFormat("{0} ", "OPTIONS");
                sb.AppendFormat("rtsp://{0}:{1}/ RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort);
                sb.AppendFormat("CSeq: {0}\r\n\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
            }
            else
            {
                sb.AppendFormat("{0} ", "OPTIONS");
                sb.AppendFormat("rtsp://{0}:{1}/ RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                sb.AppendFormat("Session: {0}\r\n\r\n", _rtspSessionId);
            }
            var result = SendRequest(sb.ToString());
            result = GetResponse();
            return result;
        }
        public bool Describe()
        {
            if ((_rtspSocket == null) )
            {
                Connect();
            }
            var sb = new StringBuilder();
            if (string.IsNullOrEmpty(_rtspSessionId))
            {
                sb.AppendFormat("{0} ", "DESCRIBE");
                sb.AppendFormat("rtsp://{0}:{1}/ RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                sb.AppendFormat("Accept: application/sdp\r\n\r\n");
            }
            else
            {
                sb.AppendFormat("{0} ", "DESCRIBE");
                sb.AppendFormat("rtsp://{0}:{1}/ RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort);
                sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
                sb.AppendFormat("Session: {0}\r\n", _rtspSessionId);
                sb.AppendFormat("Accept: application/sdp\r\n\r\n");
            }
            bool result = SendRequest(sb.ToString());
            result = GetResponse();
            /*
             * v=0
o=- 1378633020884883 1 IN IP4 192.168.2.108
s=SatIPServer:1 4
t=0 0
a=tool:idl4k
m=video 52780 RTP/AVP 33
c=IN IP4 0.0.0.0
b=AS:5000
a=control:stream=4
a=fmtp:33 ver=1.0;tuner=1,0,0,0,12344,h,dvbs2,,off,,22000,34;pids=0,100,101,102,103,106
a=sendonly

             */
            var lineArray = Regex.Split(RtspResponse, @"[\r\n]+");
            foreach (var s in lineArray)
            {
                if (s.Contains("v="))
                {
                    //var vline = Regex.Split(s, @"[=]+");
                    //Console.WriteLine(vline[1]);
                }
                if (s.Contains("o="))
                {
                    //var oline = Regex.Split(s, @"[=][-][ ]+");
                    //Console.WriteLine(oline[1] + " " + oline[2] + " " + oline[4]);
                }
                if (s.Contains("s="))
                {
                    //var sline = Regex.Split(s, @"[=][:]+");
                }
                if (s.Contains("t="))
                {
                    //var tline = Regex.Split(s, @"[=][:]+");
                }
                if (s.Contains("m="))
                {
                    //var mline = Regex.Split(s, @"[=][ ]+");
                }
                if (s.Contains("c="))
                {
                    //var cline = Regex.Split(s, @"[=][ ]+");
                }
                if (s.Contains("a="))
                {
                    //var aline = Regex.Split(s, @"[=][:][;]+");
                }

            }

            return result;
        }
        public bool TearDown()
        {
            if ((_rtspSocket == null) )
            {
                Connect();
            }
            var sb = new StringBuilder();
            sb.AppendFormat("{0} ", "TEARDOWN");
            sb.AppendFormat("rtsp://{0}:{1}/stream={2} RTSP/1.0\r\n", RtspDevice.RtspServerAddress, RtspDevice.RtspServerPort, RtspStreamId);
            sb.AppendFormat("CSeq: {0}\r\n", (++_rtspSequenceNum).ToString(CultureInfo.InvariantCulture));
            sb.AppendFormat("Session: {0}\r\n\r\n", _rtspSessionId);
            bool result = SendRequest(sb.ToString());
            Disconnect();
            return result;
        }

        #endregion

        #region Public Events

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        #region Protected Methods

        protected void OnPropertyChanged(string name)
        {
            var handler = PropertyChanged;
            if (handler != null)
            {
                handler(this, new PropertyChangedEventArgs(name));
            }
        }

        #endregion

        public void Dispose()
        {
            TearDown();
            Disconnect();
        }
    }
}
